@@ -66,6 +66,8 @@ from ..django.utils import get_field_names_by_model
 import sys
 from django.http import JsonResponse
 from .admin_env_init import DATA_UPLOAD_MAX_NUMBER_FIELDS
+from django.views.decorators.csrf import csrf_exempt
+from bdtime import tt
 
 
 IAdmin = admin.ModelAdmin
@@ -250,6 +252,8 @@ class ExportExcelMixin:
     export_fields = None        # 需要导出的字段
     extra_fields_dc = None      # 最后的verbose_name转换字典, 如: { 'ming_cheng': '哈哈哈名称' }
 
+    use_django_pandas_to_export_file_threshold = 50000  # 数据过多时则使用django-pandas生成csv临时文件, 然后返回文件流.
+
     export_asc = False      # 按id升序导出 --- 不怎么需要这个字段了
     add_index = False       # 导出时增加索引列
 
@@ -292,55 +296,80 @@ class ExportExcelMixin:
         if extra_fields_dc:
             conv_name_to_verbose_dc.update(extra_fields_dc)
 
-        response = HttpResponse(content_type='application/msexcel')
-        filename = urlquote(f"{meta.verbose_name}.xls")
-        response['Content-Disposition'] = f'attachment; filename={filename}'
-        wb = Workbook()
-        ws = wb.active
-
-        if self.export_asc:     # 不怎么需要这个字段了
-            queryset = queryset.order_by('pk')
-
-        ws.append(list(conv_name_to_verbose_dc.values()))
-
         total = queryset.count()
-        tq = tqdm(total=total)
-        for obj in queryset:
-            tq.update(1)
+        if total > self.use_django_pandas_to_export_file_threshold:
+            from django_pandas.io import read_frame
+            # qs = MyModel.objects.all()
+            df = read_frame(queryset, fieldnames=field_names)
 
-            data = []
-            for field_index in range(len(meta.fields)):
-                _meta_field = meta.fields[field_index]
-                field_name = _meta_field.name
-                if export_fields and field_name not in export_fields:
-                    continue
+            df.columns = [conv_name_to_verbose_dc.get(k) for k in df.columns]
+            df: pd.DataFrame
 
-                if isinstance(_meta_field, m.ForeignKey):
-                    # print('--- 外键: ', field_name, get_field_type_in_db(obj, field_name))
-                    to_field = _meta_field.to_fields[0]
-                    to_field = to_field if to_field else 'pk'
-                    field_value = getattr(obj, field_name)
-                    field_value = getattr(field_value, to_field)
-                else:
-                    field_value = getattr(obj, field_name)
+            temp_dir_path = os.path.join('media', 'protected_file', 'export_data')
+            os.makedirs(temp_dir_path, exist_ok=True)
+            download_file_name = f"{meta.verbose_name}.csv"
+            download_file_path = os.path.join(temp_dir_path, download_file_name)
+            # df.to_excel(download_file_path, index=False, encoding='utf-8')
+            df.to_csv(download_file_path, index=False, encoding='utf-8')
 
-                tp = get_field_type_in_py(obj, field_name)
-                if tp in ['int', 'float', 'bool']:
-                    data.append(field_value)
-                # if get_field_type_in_py(obj, field_name) == 'bool':
-                #     data.append(field_value)
-                elif tp == 'FileField':
-                    data.append(f'{field_value}')
-                else:
-                    data.append(f'{field_value}' if field_value else field_value)
+            from bddjango import DownloadFileMixin
+            response = DownloadFileMixin.download_by_file_path(download_file_path)
 
-            try:
-                ws.append(data)
-            except Exception as e:
-                print(data)
-                raise e
-        wb.save(response)
-        return response
+            # --- 清理临时文件
+            from bddjango.tools.remove_temp_file import remove_temp_file
+            remove_temp_file(temp_dir_path)
+
+            return response
+        else:
+            tq = tqdm(total=total)
+            response = HttpResponse(content_type='application/msexcel')
+            filename = urlquote(f"{meta.verbose_name}.xls")
+            response['Content-Disposition'] = f'attachment; filename={filename}'
+            wb = Workbook()
+            ws = wb.active
+
+            if self.export_asc:     # 不怎么需要这个字段了
+                queryset = queryset.order_by('pk')
+
+            ws.append(list(conv_name_to_verbose_dc.values()))
+
+            for obj in queryset:
+                tq.update(1)
+
+                data = []
+                for field_index in range(len(meta.fields)):
+                    _meta_field = meta.fields[field_index]
+                    field_name = _meta_field.name
+                    if export_fields and field_name not in export_fields:
+                        continue
+
+                    if isinstance(_meta_field, m.ForeignKey):
+                        # print('--- 外键: ', field_name, get_field_type_in_db(obj, field_name))
+                        to_field = _meta_field.to_fields[0]
+                        to_field = to_field if to_field else 'pk'
+                        field_value = getattr(obj, field_name)
+                        if field_value:
+                            field_value = getattr(field_value, to_field)
+                    else:
+                        field_value = getattr(obj, field_name)
+
+                    tp = get_field_type_in_py(obj, field_name)
+                    if tp in ['int', 'float', 'bool']:
+                        data.append(field_value)
+                    # if get_field_type_in_py(obj, field_name) == 'bool':
+                    #     data.append(field_value)
+                    elif tp == 'FileField':
+                        data.append(f'{field_value}')
+                    else:
+                        data.append(f'{field_value}' if field_value else field_value)
+
+                try:
+                    ws.append(data)
+                except Exception as e:
+                    print(data)
+                    raise e
+            wb.save(response)
+            return response
 
     export_as_excel.short_description = "导出所选数据"
     # export_as_excel.acts_on_all = True
@@ -399,6 +428,12 @@ class ImportMixin:
         super().__init__(*args, **kwargs)
         self.CHECK_EXIST_FLAG = False
         self.CREATE_EXIST_FLAG = False
+        self.COMPLETED_FLAG = False
+        self.save_tqdm_dc = {}
+        self.save_tqdm = None
+        self.check_tqdm = None
+
+        self.multi_threading_error_msg = []
 
     def import_csv(self, request):
         t_import = Time()
@@ -591,10 +626,17 @@ class ImportMixin:
                                     to_field = _meta_field.to_fields[0]
                                     to_field = to_field if to_field else 'pk'
                                     field_value = content_ls[i]
-                                    _obj_qs_ls = related_model.objects.filter(**{to_field: field_value})
-                                    assert _obj_qs_ls.exists(), f'第[{index + 2}]行外键[{verbose_i}]列找不到对应[{related_model._meta.verbose_name}_{to_field}={field_value}]的值!'
-                                    _obj = _obj_qs_ls[0]
-                                    content_ls[i] = _obj
+                                    if field_value is not None:
+                                        _obj_qs_ls = related_model.objects.filter(**{to_field: field_value})
+                                        if not _obj_qs_ls.exists():
+                                            if attr.field.db_constraint:
+                                                msg = f'第[{index + 2}]行外键[{verbose_i}]列找不到对应[{related_model._meta.verbose_name}_{to_field}={field_value}]的值!'
+                                                raise ReferenceError(msg)
+                                        _obj = _obj_qs_ls[0]
+                                        content_ls[i] = _obj
+                                    else:
+                                        content_ls[i] = None
+
                                 elif attr_field_name in ['DateField', 'DateTimeField']:
                                     if attr_field_name == 'DateTimeField':
                                         res = conv_date_time_field_str_format(ts=content_ls[i])
@@ -662,30 +704,21 @@ class ImportMixin:
                 n = df_rows
 
                 bulk_size = self.bulk_size
-                workers = self.workers
+                workers = self.workers if n > self.workers else 1
                 if workers:
                     from math import ceil
                     bulk_size = ceil(n // workers)
-                print(f'--- bulk_size: {bulk_size}, single_import_flag: {bulk_size and n < bulk_size}')
 
-                if bulk_size and n <= bulk_size:
-                    check_df_format(self, df1, md_dc_ls, curr_id, start=0, end=df_rows, n=df_rows, tqdm_i=my_tqdm, info_dc=info_dc, q=None)
-                else:
-                    a = np.arange(n)
-                    s = slice(0, len(a), bulk_size)
-                    start_end_ls = list(a[s])
-                    thread_ls = []
-                    for start_end_i in start_end_ls:
-                        thread_i = threading.Thread(target=check_df_format,
-                                                    args=(self, df1, md_dc_ls, start_end_i + curr_id, start_end_i, start_end_i + bulk_size, df_rows, my_tqdm, info_dc, q))
-                        thread_ls.append(thread_i)
-                        thread_i.start()
+                from ..django.conf import db_engine
+                use_sqlite3 = 'sqlite3' in db_engine        # sqlite3只允许单线程save, 但可以多线程check
 
-                    self.check_thread_ls(thread_ls, q, 'CHECK_EXIST_FLAG')
+                single_import_flag = n < 500 or (bulk_size and n <= bulk_size)      # 是否单线程导入
+                multi_single_import_flag = bulk_size and n <= bulk_size         # 是否先返回然后再开另一个线程处理数据
+
+                print(f'--- workers: {workers} --- bulk_size: {bulk_size}, single_import_flag: {single_import_flag}')
 
                 save_tqdm = tqdm(total=df_rows)
                 save_tqdm.desc = '写入数据库ing...'
-                print(f'*** df_rows: {df_rows} --- len(md_dc_ls): {len(md_dc_ls)}')
                 save_q = queue.Queue()
 
                 def save_md_dc_ls(md_dc_ls, start=0, end=None, n=None, tqdm_i=None, q=None, pk_name=None):
@@ -699,6 +732,10 @@ class ImportMixin:
                         end = n
                     if end > n:
                         end = n
+
+                    i = start
+                    md_dc_i = {}
+
                     try:
                         for i in range(start, end):
                             # print(start_end_i, start_end_i + bulk_size)
@@ -725,41 +762,120 @@ class ImportMixin:
                         else:
                             raise e
 
-                from ..django.conf import db_engine
-                use_sqlite3 = 'sqlite3' in db_engine
-                if (bulk_size and n <= bulk_size) or use_sqlite3:
+                self.check_tqdm = my_tqdm
+                self.COMPLETED_FLAG = False
+                self.CHECK_EXIST_FLAG = False
+                self.CREATE_EXIST_FLAG = False
+                self.multi_threading_error_msg = []
+
+                if single_import_flag:
+                    check_df_format(self, df1, md_dc_ls, curr_id, start=0, end=df_rows, n=df_rows, tqdm_i=my_tqdm, info_dc=info_dc, q=None)
+
                     for i in range(len(md_dc_ls)):
                         md_dc_i = md_dc_ls[i]
                         save_tqdm.update(1)
-                        self.model.objects.create(**md_dc_i)                 # `md_i.save()`不太好, 容易覆盖原有数据
+                        self.model.objects.create(**md_dc_i)  # `md_i.save()`不太好, 容易覆盖原有数据
                     msg = f"{f_format}文件导入成功! 一共导入{df_rows}条数据, 耗时: {t_import.now(1)}秒."
+
+                    reset_db_sequence(self.model)
+                    # msg = f'成功导入{n}条数据!'
+                    self.message_user(request, msg)
+                    self.remove_temp_file(tempdir)
+
+                    return redirect("..")
                 else:
-                    a = np.arange(n)
-                    s = slice(0, len(a), bulk_size)
-                    start_end_ls = list(a[s])
-                    thread_ls = []
-                    for start_end_i in start_end_ls:
-                        # print(start_end_i, start_end_i + bulk_size)
-                        thread_i = threading.Thread(target=save_md_dc_ls, args=(md_dc_ls, start_end_i, start_end_i + bulk_size, n, save_tqdm, save_q, pk_name))
-                        thread_ls.append(thread_i)
-                        thread_i.start()
-                        thread_i.join()
+                    # self.check_tqdm = my_tqdm
 
-                    # self.CREATE_EXIST_FLAG = False
-                    self.check_thread_ls(thread_ls, save_q, 'CREATE_EXIST_FLAG')
+                    def multi_threading_importer():
+                        if multi_single_import_flag:
+                            check_df_format(self, df1, md_dc_ls, curr_id, start=0, end=df_rows, n=df_rows, tqdm_i=my_tqdm,
+                                        info_dc=info_dc, q=None)
+                        else:
+                            a = np.arange(n)
+                            s = slice(0, len(a), bulk_size)
+                            start_end_ls = list(a[s])
+                            thread_ls = []
+                            for start_end_i in start_end_ls:
+                                thread_i = threading.Thread(target=check_df_format,
+                                                            args=(
+                                                            self, df1, md_dc_ls, start_end_i + curr_id, start_end_i,
+                                                            start_end_i + bulk_size, df_rows, my_tqdm, info_dc, q))
+                                thread_ls.append(thread_i)
+                                thread_i.start()
 
-                    # for thread_i in thread_ls:
-                    #     thread_i.join()
-                    msg = f"{f_format}文件导入中...<br>一共将导入{df_rows}条数据, <br>请稍后刷新查看是否完全导入."
+                            self.check_thread_ls(thread_ls, q, 'CHECK_EXIST_FLAG')
 
-                reset_db_sequence(self.model)
-                self.message_user(request, msg)
-                self.remove_temp_file(tempdir)
+                        self.check_tqdm = None
+                        self.save_tqdm = save_tqdm
+
+                        if multi_single_import_flag or use_sqlite3:
+                            for i in range(len(md_dc_ls)):
+                                if self.CREATE_EXIST_FLAG:
+                                    break
+
+                                md_dc_i = md_dc_ls[i]
+                                save_tqdm.update(1)
+                                self.model.objects.create(**md_dc_i)  # `md_i.save()`不太好, 容易覆盖原有数据
+                        else:
+                            # t_import.tqdm_sleep(f'=== 将使用{workers}个workers导入{n}条数据, 每个worker导入{bulk_size}条!')
+                            a = np.arange(n)
+                            s = slice(0, len(a), bulk_size)
+                            start_end_ls = list(a[s])
+                            thread_ls = []
+                            for start_end_i in start_end_ls:
+                                # print(start_end_i, start_end_i + bulk_size)
+                                thread_i = threading.Thread(target=save_md_dc_ls, args=(
+                                    md_dc_ls, start_end_i, start_end_i + bulk_size, n, save_tqdm, save_q, pk_name))
+                                thread_ls.append(thread_i)
+                                thread_i.start()
+
+                            self.check_thread_ls(thread_ls, save_q, 'CREATE_EXIST_FLAG')
+
+                        self.save_tqdm = None
+                        self.COMPLETED_FLAG = True
+
+                    thread_importer = threading.Thread(target=multi_threading_importer)
+                    thread_importer.start()
+
+                    meta = self.model._meta
+                    info = f'{self.model._meta.app_label}.{self.model._meta.object_name}'
+                    title = f'{self.model._meta.verbose_name}'
+
+                    href_str = f"/api/admin/{meta.app_label}/{meta.model_name}/get_import_info/?info={info}&title={title}"
+                    a_str = f'<a href="{href_str}" target="_blank">点击此处查看进度</a>'
+                    dst_number = df_rows + self.model.objects.count()
+
+                    w_str = f"将使用[{workers}]个线程" if workers > 1 else ""
+                    msg = f"[{meta.verbose_name}]的{f_format}文件导入中...<br>将{w_str}导入[{df_rows}]条数据, <br>请在导入完毕后查看是否为[{dst_number}]条数据. <br>{a_str}"
+                    self.message_user(request, msg)
+                    return redirect(href_str)
+
+                # if not use_progress_bar_flag:
+                #     reset_db_sequence(self.model)
+                #     self.message_user(request, msg)
+                #     self.remove_temp_file(tempdir)
+                #
+                #     return redirect("..")
+                # else:
+                #
+                #     meta = self.model._meta
+                #     info = f'{self.model._meta.app_label}.{self.model._meta.object_name}'
+                #
+                #     href_str = f"/api/admin/{meta.app_label}/{meta.model_name}/get_import_info/?info={info}"
+                #     a_str = f'<a href="{href_str}" target="_blank">点击此处查看进度</a>'
+                #     dst_number = df_rows + self.model.objects.count()
+                #     msg = f"{f_format}文件导入中...<br>一共将导入[{df_rows}]条数据, <br>请在导入完毕后查看是否为[{dst_number}]条数据. <br>{a_str}"
+                #     self.message_user(request, msg)
+                #     return redirect(href_str)
 
-                return redirect("..")
         except Exception as e:
             msg = f"数据导入失败!</br>错误信息：&nbsp;" + str(e)
 
+            self.multi_threading_error_msg.append(msg)
+            self.save_tqdm = None
+            self.check_tqdm = None
+            tt.tqdm_sleep(f'****** multi_threading_error_msg: {msg}')
+
             ret_restful_api = request._post.get('ret_restful_api')  # 是否返回RestfulAPI格式
 
             from ..pure import convert_query_parameter_to_bool
@@ -786,6 +902,106 @@ class ImportMixin:
         payload = {"form": form}
         return render(request, "admin/csv_form.html", payload)
 
+    def _get_import_info(self, request):
+        # print(self.save_tqdm)
+        # from math import ceil
+
+        def get_data_from_tqdm(my_tqdm):
+            dc = my_tqdm.__dict__
+
+            n = dc.get('n')
+            total = dc.get('total')
+            desc = dc.get('desc')
+
+            tq_str = str(my_tqdm)
+            t_str = ""
+            reg = re.compile(r'\[.*\]')
+            match = reg.search(tq_str)
+            if match:
+                t_str = match.group(0)
+
+            # _progress = n * 100 / total
+            # progress = ceil(_progress) if _progress >= 98 else int(_progress)
+
+            progress = int(n * 100 / total)
+
+            ret = {
+                'desc': desc,
+                'n': n,
+                'total': total,
+                'progress': progress,
+                "t_str": t_str
+            }
+            return ret
+
+        tt = Time()
+        if len(self.multi_threading_error_msg):
+            error_msg = self.multi_threading_error_msg.pop(0)
+            return JsonResponse(
+                data={
+                    'status': 404,
+                    'msg': 'error',
+                    'result': error_msg
+                }
+            )
+
+        if self.save_tqdm is None and self.check_tqdm is None:
+            if self.COMPLETED_FLAG:
+                status = 201
+            else:
+                status = 204
+            return JsonResponse(
+                data={
+                    'status': status,
+                    'msg': f'{tt.get_current_beijing_time_str(decimal_places=1)} --- 当前没有正在导入的数据!',
+                    'result': None
+                }
+            )
+        else:
+
+            if self.check_tqdm is not None:
+                res = get_data_from_tqdm(self.check_tqdm)
+            else:
+                res = get_data_from_tqdm(self.save_tqdm)
+            return JsonResponse(
+                data={
+                    'status': 200,
+                    'msg': 'ok',
+                    'result': res
+                }
+            )
+
+    def get_import_info(self, request):
+        # from bddjango import APIResponse
+        # return APIResponse(111)
+
+        # payload = {}
+        # return render(request, "admin/import_progress_bar.html", payload)
+        # return 1
+
+        from django.http import FileResponse
+
+        from bddjango import get_root_path
+        bd_root = get_root_path()
+        bd_template_admin_dir = os.path.join(bd_root, "templates", "admin")
+        file_path = os.path.join(bd_template_admin_dir, 'import_progress_bar.html')
+        file = open('bddjango/templates/admin/import_progress_bar.html', 'rb')
+        response = FileResponse(file)
+        response['Content-Type'] = 'text/html'
+        return response
+
+    def stop_import(self, request):
+        self.CHECK_EXIST_FLAG = True
+        self.CREATE_EXIST_FLAG = True
+        self.COMPLETED_FLAG = True
+        return JsonResponse(
+            data={
+                'status': 200,
+                'msg': 'ok',
+                'result': None
+            }
+        )
+
     def check_thread_ls(self, thread_ls, q, flag_attr_name):
         from bdtime import tt
         while True:
@@ -806,8 +1022,16 @@ class ImportMixin:
                 setattr(self, flag_attr_name, True)
 
                 try:
-                    raise Exception(f'{last[-2]}' if len(last) >= 2 else f'{last[-1]}')
+                    msg = f'{last[-2]}' if len(last) >= 2 else f'{last[-1]}'
+                    self.multi_threading_error_msg.append(msg)
+                    self.save_tqdm = None
+                    self.check_tqdm = None
+                    raise Exception(msg)
                 except:
+                    msg = f'{last}'
+                    self.multi_threading_error_msg.append(msg)
+                    self.save_tqdm = None
+                    self.check_tqdm = None
                     raise Exception(last)
         print(u'end')
 
@@ -844,10 +1068,20 @@ class ImportMixin:
         return ret
 
     def get_urls(self):
+        from django.views.static import serve
+        from bddjango import get_root_path
+        bd_root = get_root_path()
+        bd_template_admin_dir = os.path.join(bd_root, "templates", "admin")
+
         my_urls = [
             path('import-csv/', self.import_csv),
             path('export_all_csv/', self.export_all_csv),
             path('export_all_excel/', self.export_all_excel),
+            # path('get_import_info/', self.get_import_info),
+            path('get_import_info/', serve, {'document_root': bd_template_admin_dir, 'path': 'import_progress_bar.html'}),
+            # path('get_import_info/', self.get_import_info),
+            path('_get_import_info/', self._get_import_info),
+            path('stop_import/', self.stop_import),
         ]
         return my_urls + super().get_urls()
 
