@@ -427,6 +427,9 @@ class AutoWiki(APIView):
             for field_name, verbose_name, can_be_empty, field_type in zip(field_names, verbose_names, can_be_empty_ls, field_type_ls):
                 print(field_name, verbose_name, can_be_empty)
 
+                if verbose_name == 'search_rank':
+                    verbose_name = "检索排序字段"
+
                 si = ''
                 if sf == '__all__' or field_name in sf:
                     # field_type = get_field_type_in_py(md, field_name)
