@@ -141,4 +141,80 @@ class ListStatisticMixin:
         return ret, status, msg
 
 
+class DownloadFileMixin:
+    """
+    下载文件(若目标为文件夹, 则打包为zip下载)
+    """
+
+    default_path = None
+    path_field_name = None
+
+    @staticmethod
+    def _file_iterator(file_name, open_model='rb', chunk_size=512):
+        with open(file_name, open_model) as f:  # , encoding=encoding
+            while True:
+                c = f.read(chunk_size)
+                if c:
+                    yield c
+                else:
+                    break
+
+    @staticmethod
+    def download_by_file_path(download_file_path):
+        # --- 返回文件流
+        from django.http import StreamingHttpResponse
+        from django.utils.http import urlquote
+        import os
+
+        download_file_name = os.path.basename(download_file_path)
+        response = StreamingHttpResponse(DownloadFileMixin._file_iterator(download_file_path))
+        # response['Content-Type'] = 'application/octet-stream'
+        filename = urlquote(download_file_name)
+        response['Content-Type'] = 'application/x-zip-compressed'
+        response['Content-Disposition'] = 'attachment;filename="{0}"'.format(filename)
+        return response
+
+    def download_file(self, request, queryset=None, model=None):
+        import os
+
+        assert queryset.count() == 1, '一次只允许下载一个文件!'
+        obj = queryset[0]
+
+        src = getattr(obj, self.path_field_name) if self.path_field_name else self.default_path
+        assert src, '目标路径不能为空! 请指定`path_field_name`或者`default_path`!'
+        assert os.path.exists(src), f'目标路径[{src}]不存在!'
+
+        if os.path.isdir(src):
+            # --- 创建临时路径
+            from bddjango import get_base_model
+
+            md = get_base_model(obj)
+            tmp_file_dir_path = f"tempdir/{md._meta.app_label}/{md._meta.object_name}"
+            os.makedirs(tmp_file_dir_path, exist_ok=True)
+
+            # --- 打包为zip
+            from bddjango.tools.zipDir import zipDir
+
+            zip_file_name = f'{obj}.zip'
+            dst = os.path.join(tmp_file_dir_path, zip_file_name)
+
+            zipDir(src, dst)
+
+            download_file_path = dst
+
+            # --- 清理临时文件
+            from bddjango.tools.remove_temp_file import remove_temp_file
+            remove_temp_file(tmp_file_dir_path)
+        else:
+            download_file_path = src
+
+        response = DownloadFileMixin.download_by_file_path(download_file_path)
+        return response
+
+    download_file.short_description = '下载文件'
+    # download_file.type = 'warning'
+    download_file.icon = 'el-icon-download'
+    download_file.enable = True
+    download_file.confirm = f"确定{download_file.short_description}?"
+
 
