@@ -252,6 +252,12 @@ def conv_queryset_ls_to_serialzer_ls(qs_ls: list):
     return dc_ls
 
 
+def conv_qs_ls_to_serializer_data(qs_ls):
+    base_serializer = get_base_serializer(qs_ls)
+    dc_ls = base_serializer(qs_ls, many=True).data
+    return dc_ls
+
+
 def get_field_type_in_db(model, field_name):
     """根据模型和字段名, 获取这个字段在数据库中对应的类型"""
     tp = model._meta.get_field(field_name).get_internal_type()
@@ -467,12 +473,14 @@ def order_by_order_type_ls(queryset, order_type_ls) -> QuerySet:
     else:
         ls = []
         for order_type in order_type_ls:
-            if order_type:
+            if isinstance(order_type, str):
                 if order_type.startswith('-'):
                     order_type1 = order_type[1:]
                     ls.append(F(order_type1).desc(nulls_last=True))
                 else:
                     ls.append(F(order_type).asc(nulls_first=True))
+            else:
+                ls.append(order_type)
         ret = queryset.order_by(*ls)
     return ret
 
@@ -773,59 +781,98 @@ def judge_db_is_migrating():
         return False
 
 
-def get_total_occurrence_times_by_keywords(total_qs_ls, search_field_ls, keywords, get_frequence=False):
+def get_total_occurrence_times_by_keywords(total_qs_ls, search_field_ls, keywords, get_frequence=True, topK=5, rank_field_name=None, rank__gte=None, search_weight_ls=None):
     """
-    统计keywords在qs_ls的search_field_ls中是否出现
+    # 关键词次数统计
+    - 统计keywords在qs_ls的search_field_ls中是否出现, 以及出现次数.
+    - 可以出现次数作为相关性排序依据
 
     :param total_qs_ls: 用来统计的queryset
     :param search_field_ls: 要匹配的字段
-    :param kw: 用来检索的关键词
-    :param get_frequence: 是否精确计算kw在字段中出现的次数
-    :return: queryset, 且annotate出现次数, 存在`total_occurrence_times`字段中
+    :param search_weight_ls: 字段对应权重
+    :param keywords: 用来检索的关键词
+    :param get_frequence: 是否精确计算`keywords`在字段中出现的次数
+    :param topK: 提取关键词的个数
+    :param rank_field_name: annotate出来的排序字段名, 默认`searh_rank`
+    :param rank__gte: 出现次数过滤的最小阈值
+    :return: queryset, 且annotate出现次数, 存在`rank_field_name`字段中
     """
+    from django.db import models as m
     from django.db.models import functions
 
-    my_api_assert_function(isinstance(keywords, list), 'keywords的类型必须为list!')
+    rank_field_name = rank_field_name if rank_field_name else 'search_rank'
+    rank__gte = rank__gte if rank__gte is not None else 0.01
+
+    if isinstance(keywords, str):
+        from bddjango.tools.extract_keyword import extract_keywords
+        keywords = extract_keywords.handle(keywords, cut_all=False, topK=topK)
+
+    assert isinstance(keywords, list), 'keywords的类型必须为str或list!'
+
     search_dc = {}
-    occurrence_times_ls = []
+    occurance_times_ls = []
 
     for k in range(len(keywords)):
+        # k = 1
         kw = keywords[k]
         kw_l = len(kw)
         for sf_i in search_field_ls:
             if get_frequence:
-                k_name = f'k{k}_{sf_i}_occurrence_times'
+                k_name = f'k{k}_in_{sf_i}'
+
+                # 统计每个`keyword`的出现次数
                 dc = {
                     k_name: (functions.Length(sf_i) - functions.Length(
-                        functions.Replace(sf_i, m.Value(kw), m.Value('')))) / kw_l  # 出现次数
+                        functions.Replace(sf_i, m.Value(kw), m.Value('')))) / kw_l
                 }
             else:
                 k_name = f'k{k}_in_{sf_i}'
                 dc = {
-                    k_name: m.Exists(total_qs_ls.filter(id=m.OuterRef('id')).filter(**{f'{sf_i}__contains': kw})),   # 判断是否在title中
+                    k_name: m.Exists(total_qs_ls.filter(pk=m.OuterRef('pk')).filter(**{f'{sf_i}__contains': kw})),   # 判断是否在title中
                 }
-            occurrence_times_ls.append(k_name)
+            occurance_times_ls.append(k_name)
             search_dc.update(dc)
     res_qs_ls: m.QuerySet = total_qs_ls.annotate(**search_dc)
+
+    # from bddjango import show_json, show_ls
     # show_ls(res_qs_ls.values(*(['id'] + search_field_ls + list(search_dc.keys())))[:3])
+    # print(f'--- 检索字段: {search_field_ls}')
+    # show_ls(res_qs_ls.values(*(['id'] + list(search_dc.keys())))[:3])
 
     f_ls = 0
-    for i in occurrence_times_ls:
+    for i in range(len(occurance_times_ls)):
+        # i = 1
+        k_name = occurance_times_ls[i]
+
+        if search_weight_ls:
+            assert len(search_weight_ls) == len(search_field_ls), '`search_weight_ls`和`search_field_ls`长度不一致!'
+            keyword_i = k_name.split('_in_')[-1]
+            _i = search_field_ls.index(keyword_i)
+            weight_i = search_weight_ls[_i]
+            _then = m.F(k_name) * weight_i if get_frequence else m.Value(weight_i)
+        else:
+            _then = m.F(k_name) if get_frequence else m.Value(1)
+
         if get_frequence:
-            f_ls += F(i)
+            # f_ls += m.F(i)
+            f_ls += m.Case(
+                m.When(**{f'{k_name}__isnull': False}, then=_then),
+                default=m.Value(0),
+                output_field=m.FloatField()
+            )
         else:
             f_ls += m.Case(
-                m.When(**{i: True}, then=m.Value(1)),
+                m.When(**{k_name: True}, then=_then),
                 default=0,
-                output_field=m.IntegerField()
+                output_field=m.FloatField()
             )
 
-    res_qs_ls = res_qs_ls.annotate(**{'total_occurrence_times': f_ls})
-    ret = res_qs_ls
+    ret_qs_ls = res_qs_ls.annotate(**{rank_field_name: f_ls})
+    ret_qs_ls = ret_qs_ls.filter(**{f'{rank_field_name}__gte': rank__gte}).order_by(*[f'-{rank_field_name}', 'pk'])
+    ret = ret_qs_ls
     return ret
 
 
-
 def get_statistic_fields_result(queryset, statistic_fields, statistic_size=5, descend=1, order_config_dc_ls=None):
     """
     # 统计字段 fields 的值各出现了几次
@@ -838,10 +885,10 @@ def get_statistic_fields_result(queryset, statistic_fields, statistic_size=5, de
     assert isinstance(statistic_fields, list), 'statistic_fields必须为list类型!'
     statistic_size = int(statistic_size)
 
-    if descend:
-        ordering = ['-counts']
-    else:
-        ordering = ['counts']
+    # if descend:
+    #     ordering = ['-counts']
+    # else:
+    #     ordering = ['counts']
 
     if descend == 0:
         ordering = ['counts']
@@ -851,6 +898,8 @@ def get_statistic_fields_result(queryset, statistic_fields, statistic_size=5, de
         ordering = descend
     elif descend == 3:  # self倒序
         ordering = descend
+    else:
+        raise ValueError('ordering必须在[0, 3]之间, 取值: {0: `counts`顺序, 1:`counts`倒叙, 2: `self`顺序, 3: `self`倒叙}')
 
     order_df = None
     if order_config_dc_ls:
@@ -870,7 +919,6 @@ def get_statistic_fields_result(queryset, statistic_fields, statistic_size=5, de
     for field in statistic_fields:
         # break
         if isinstance(field, (tuple, list)):
-            # field_qsv = queryset.values(*field)
             field_qsv = queryset.values(field[0])
             dc_name = field[0]
         else:
@@ -880,6 +928,7 @@ def get_statistic_fields_result(queryset, statistic_fields, statistic_size=5, de
         if not isinstance(ordering, list):
             assert ordering in [2, 3], 'descend为2, 则按自身顺序排序, 为3则按自身倒序排序'
             ordering = dc_name if ordering == 2 else f'-{dc_name}'
+            ordering = [ordering]
 
         if order_df is not None and dc_name in order_df.index:
             _ordering = order_df.loc[dc_name][0]
