@@ -306,4 +306,8 @@ def find_indexes_in_dc_ls(dc_ls, key, value, find_all=False):
     return indexes
 
 
+from .tools.remove_path import remove_path
+from .tools.copy_to import copy_to
+from .tools.unzip_to import unzip_to
+
 
